@@ -17,12 +17,30 @@
 
 #include "private.h"
 
+#if LUV_UV_VERSION_GEQ(1, 28, 0)
+typedef struct {
+  uv_dir_t* handle;
+  int dirents_ref; /* handle has been closed if this is LUA_NOREF */
+} luv_dir_t;
+#endif
+
 static uv_fs_t* luv_check_fs(lua_State* L, int index) {
+  if (luaL_testudata(L, index, "uv_fs") != NULL) {
+    return (uv_fs_t*)lua_touserdata(L, index);
+  }
   uv_fs_t* req = (uv_fs_t*)luaL_checkudata(L, index, "uv_req");
-  luaL_argcheck(L, req->type = (uv_req_type)(UV_FS && req->data), index, "Expected uv_fs_t");
+  luaL_argcheck(L, req->type == UV_FS && req->data, index, "Expected uv_fs_t");
   return req;
 }
 
+static int luv_fs_gc(lua_State* L) {
+  uv_fs_t* req = luv_check_fs(L, 1);
+  luv_cleanup_req(L, (luv_req_t*)req->data);
+  req->data = NULL;
+  uv_fs_req_cleanup(req);
+  return 0;
+}
+
 static void luv_push_timespec_table(lua_State* L, const uv_timespec_t* t) {
   lua_createtable(L, 0, 2);
   lua_pushinteger(L, t->tv_sec);
@@ -208,6 +226,20 @@ static void luv_push_statfs_table(lua_State* L, const uv_statfs_t* s) {
 };
 #endif
 
+static int fs_req_has_dest_path(uv_fs_t* req) {
+  switch (req->fs_type) {
+    case UV_FS_RENAME:
+    case UV_FS_SYMLINK:
+    case UV_FS_LINK:
+#if LUV_UV_VERSION_GEQ(1, 14, 0)
+    case UV_FS_COPYFILE:
+#endif
+      return 1;
+    default:
+      return 0;
+  }
+}
+
 /* Processes a result and pushes the data onto the stack
    returns the number of items pushed */
 static int push_fs_result(lua_State* L, uv_fs_t* req) {
@@ -220,7 +252,13 @@ static int push_fs_result(lua_State* L, uv_fs_t* req) {
 
   if (req->result < 0) {
     lua_pushnil(L);
-    if (req->path) {
+    if (fs_req_has_dest_path(req)) {
+      lua_rawgeti(L, LUA_REGISTRYINDEX, data->data_ref);
+      const char* dest_path = lua_tostring(L, -1);
+      lua_pop(L, 1);
+      lua_pushfstring(L, "%s: %s: %s -> %s", uv_err_name(req->result), uv_strerror(req->result), req->path, dest_path);
+    }
+    else if (req->path) {
       lua_pushfstring(L, "%s: %s: %s", uv_err_name(req->result), uv_strerror(req->result), req->path);
     }
     else {
@@ -313,14 +351,15 @@ static int push_fs_result(lua_State* L, uv_fs_t* req) {
       luaL_unref(L, LUA_REGISTRYINDEX, data->data_ref);
       data->data_ref = LUA_NOREF;
 
-      (*(void**)lua_newuserdata(L, sizeof(void*))) = dir;
-      lua_pushfstring(L, "uv_dir:%p", dir);
-      dir->dirents = lua_newuserdata(L, sizeof(uv_dirent_t)*nentries);
-      dir->nentries = nentries;
-      lua_rawset(L, LUA_REGISTRYINDEX);
+      luv_dir_t* luv_dir = lua_newuserdata(L, sizeof(*luv_dir));
       luaL_getmetatable(L, "uv_dir");
       lua_setmetatable(L, -2);
 
+      luv_dir->handle = dir;
+      luv_dir->handle->dirents = lua_newuserdata(L, sizeof(uv_dirent_t)*nentries);
+      luv_dir->dirents_ref = luaL_ref(L, LUA_REGISTRYINDEX);
+      luv_dir->handle->nentries = nentries;
+
       return 1;
     }
     case UV_FS_READDIR: {
@@ -352,6 +391,11 @@ static int push_fs_result(lua_State* L, uv_fs_t* req) {
 
 static void luv_fs_cb(uv_fs_t* req) {
   luv_req_t* data = (luv_req_t*)req->data;
+  // This can happen if luv_fs_cb is called during loop gc. For example, this can happen
+  // when the async version of fs_scandir is called but the loop is never run before the process exits.
+  //
+  // TODO: A more comprehensive fix for this problem would be related to https://github.com/luvit/luv/issues/437
+  if (data == NULL) return;
   lua_State* L = data->ctx->L;
 
   int nargs = push_fs_result(L, req);
@@ -390,7 +434,16 @@ static void luv_fs_cb(uv_fs_t* req) {
                      sync ? NULL : luv_fs_cb);            \
   if (req->fs_type != UV_FS_ACCESS && ret < 0) {          \
     lua_pushnil(L);                                       \
-    if (req->path) {                                      \
+    if (fs_req_has_dest_path(req)) {                      \
+      lua_rawgeti(L, LUA_REGISTRYINDEX, data->data_ref);  \
+      const char* dest_path = lua_tostring(L, -1);        \
+      lua_pop(L, 1);                                      \
+      lua_pushfstring(L, "%s: %s: %s -> %s",              \
+          uv_err_name(req->result),                       \
+          uv_strerror(req->result),                       \
+          req->path, dest_path);                          \
+    }                                                     \
+    else if (req->path) {                                 \
       lua_pushfstring(L, "%s: %s: %s",                    \
           uv_err_name(req->result),                       \
           uv_strerror(req->result), req->path);           \
@@ -401,9 +454,11 @@ static void luv_fs_cb(uv_fs_t* req) {
           uv_strerror(req->result));                      \
     }                                                     \
     lua_pushstring(L, uv_err_name(req->result));          \
-    luv_cleanup_req(L, data);                             \
-    req->data = NULL;                                     \
-    uv_fs_req_cleanup(req);                               \
+    if(req->fs_type != UV_FS_SCANDIR) {                   \
+      luv_cleanup_req(L, data);                           \
+      req->data = NULL;                                   \
+      uv_fs_req_cleanup(req);                             \
+    }                                                     \
     nargs = 3;                                            \
   }                                                       \
   else if (sync) {                                        \
@@ -430,7 +485,7 @@ static int luv_fs_close(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(close, req, file);
 }
@@ -441,7 +496,7 @@ static int luv_fs_open(lua_State* L) {
   int flags = luv_check_flags(L, 2);
   int mode = luaL_checkinteger(L, 3);
   int ref = luv_check_continuation(L, 4);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(open, req, path, flags, mode);
 }
@@ -450,15 +505,24 @@ static int luv_fs_read(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   int64_t len = luaL_checkinteger(L, 2);
-  int64_t offset = luaL_checkinteger(L, 3);
-  uv_buf_t buf;
+  // -1 offset means "the current file offset is used and updated"
+  int64_t offset = -1;
   int ref;
-  uv_fs_t* req;
+  // both offset and callback are optional
+  if (luv_is_callable(L, 3) && lua_isnoneornil(L, 4)) {
+    ref = luv_check_continuation(L, 3);
+  }
+  else {
+    offset = luaL_optinteger(L, 3, offset);
+    ref = luv_check_continuation(L, 4);
+  }
   char* data = (char*)malloc(len);
-  if (!data) return luaL_error(L, "Failure to allocate buffer");
-  buf = uv_buf_init(data, len);
-  ref = luv_check_continuation(L, 4);
-  req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  if (!data) {
+    luaL_unref(L, LUA_REGISTRYINDEX, ref);
+    return luaL_error(L, "Failure to allocate buffer");
+  }
+  uv_buf_t buf = uv_buf_init(data, len);
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   // TODO: find out why we can't just use req->ptr for the base
   ((luv_req_t*)req->data)->data = buf.base;
@@ -469,7 +533,7 @@ static int luv_fs_unlink(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(unlink, req, path);
 }
@@ -477,9 +541,18 @@ static int luv_fs_unlink(lua_State* L) {
 static int luv_fs_write(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
-  int64_t offset = luaL_checkinteger(L, 3);
-  int ref = luv_check_continuation(L, 4);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  // -1 offset means "the current file offset is used and updated"
+  int64_t offset = -1;
+  int ref;
+  // both offset and callback are optional
+  if (luv_is_callable(L, 3) && lua_isnoneornil(L, 4)) {
+    ref = luv_check_continuation(L, 3);
+  }
+  else {
+    offset = luaL_optinteger(L, 3, offset);
+    ref = luv_check_continuation(L, 4);
+  }
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   size_t count;
   uv_buf_t* bufs = luv_check_bufs(L, 2, &count, (luv_req_t*)req->data);
@@ -494,7 +567,7 @@ static int luv_fs_mkdir(lua_State* L) {
   const char* path = luaL_checkstring(L, 1);
   int mode = luaL_checkinteger(L, 2);
   int ref = luv_check_continuation(L, 3);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(mkdir, req, path, mode);
 }
@@ -503,7 +576,7 @@ static int luv_fs_mkdtemp(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   const char* tpl = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(mkdtemp, req, tpl);
 }
@@ -513,7 +586,7 @@ static int luv_fs_mkstemp(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   const char* tpl = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(mkstemp, req, tpl);
 }
@@ -523,7 +596,7 @@ static int luv_fs_rmdir(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(rmdir, req, path);
 }
@@ -533,8 +606,8 @@ static int luv_fs_scandir(lua_State* L) {
   const char* path = luaL_checkstring(L, 1);
   int flags = 0; // TODO: find out what these flags are.
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ctx, ref);
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
+  req->data = luv_setup_req_with_mt(L, ctx, ref, "uv_fs");
   FS_CALL(scandir, req, path, flags);
 }
 
@@ -542,12 +615,7 @@ static int luv_fs_scandir_next(lua_State* L) {
   uv_fs_t* req = luv_check_fs(L, 1);
   uv_dirent_t ent;
   int ret = uv_fs_scandir_next(req, &ent);
-  if (ret == UV_EOF) {
-    luv_cleanup_req(L, (luv_req_t*)req->data);
-    req->data = NULL;
-    uv_fs_req_cleanup(req);
-    return 0;
-  }
+  if (ret == UV_EOF) return 0;
   if (ret < 0) return luv_error(L, ret);
   return luv_push_dirent(L, &ent, 0);
 }
@@ -556,7 +624,7 @@ static int luv_fs_stat(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(stat, req, path);
 }
@@ -565,7 +633,7 @@ static int luv_fs_fstat(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(fstat, req, file);
 }
@@ -574,7 +642,7 @@ static int luv_fs_lstat(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(lstat, req, path);
 }
@@ -584,8 +652,11 @@ static int luv_fs_rename(lua_State* L) {
   const char* path = luaL_checkstring(L, 1);
   const char* new_path = luaL_checkstring(L, 2);
   int ref = luv_check_continuation(L, 3);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
+  // ref the dest path so that we can print it in the error message
+  lua_pushvalue(L, 2);
+  ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
   FS_CALL(rename, req, path, new_path);
 }
 
@@ -593,7 +664,7 @@ static int luv_fs_fsync(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(fsync, req, file);
 }
@@ -602,7 +673,7 @@ static int luv_fs_fdatasync(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(fdatasync, req, file);
 }
@@ -612,7 +683,7 @@ static int luv_fs_ftruncate(lua_State* L) {
   uv_file file = luaL_checkinteger(L, 1);
   int64_t offset = luaL_checkinteger(L, 2);
   int ref = luv_check_continuation(L, 3);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(ftruncate, req, file, offset);
 }
@@ -624,7 +695,7 @@ static int luv_fs_sendfile(lua_State* L) {
   int64_t in_offset = luaL_checkinteger(L, 3);
   size_t length = luaL_checkinteger(L, 4);
   int ref = luv_check_continuation(L, 5);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(sendfile, req, out_fd, in_fd, in_offset, length);
 }
@@ -634,7 +705,7 @@ static int luv_fs_access(lua_State* L) {
   const char* path = luaL_checkstring(L, 1);
   int amode = luv_check_amode(L, 2);
   int ref = luv_check_continuation(L, 3);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(access, req, path, amode);
 }
@@ -644,7 +715,7 @@ static int luv_fs_chmod(lua_State* L) {
   const char* path = luaL_checkstring(L, 1);
   int mode = luaL_checkinteger(L, 2);
   int ref = luv_check_continuation(L, 3);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(chmod, req, path, mode);
 }
@@ -654,7 +725,7 @@ static int luv_fs_fchmod(lua_State* L) {
   uv_file file = luaL_checkinteger(L, 1);
   int mode = luaL_checkinteger(L, 2);
   int ref = luv_check_continuation(L, 3);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(fchmod, req, file, mode);
 }
@@ -665,7 +736,7 @@ static int luv_fs_utime(lua_State* L) {
   double atime = luaL_checknumber(L, 2);
   double mtime = luaL_checknumber(L, 3);
   int ref = luv_check_continuation(L, 4);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(utime, req, path, atime, mtime);
 }
@@ -676,7 +747,7 @@ static int luv_fs_futime(lua_State* L) {
   double atime = luaL_checknumber(L, 2);
   double mtime = luaL_checknumber(L, 3);
   int ref = luv_check_continuation(L, 4);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(futime, req, file, atime, mtime);
 }
@@ -688,7 +759,7 @@ static int luv_fs_lutime(lua_State* L) {
   double atime = luaL_checknumber(L, 2);
   double mtime = luaL_checknumber(L, 3);
   int ref = luv_check_continuation(L, 4);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(lutime, req, path, atime, mtime);
 }
@@ -699,8 +770,11 @@ static int luv_fs_link(lua_State* L) {
   const char* path = luaL_checkstring(L, 1);
   const char* new_path = luaL_checkstring(L, 2);
   int ref = luv_check_continuation(L, 3);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
+  // ref the dest path so that we can print it in the error message
+  lua_pushvalue(L, 2);
+  ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
   FS_CALL(link, req, path, new_path);
 }
 
@@ -710,18 +784,31 @@ static int luv_fs_symlink(lua_State* L) {
   const char* new_path = luaL_checkstring(L, 2);
   int flags = 0, ref;
   uv_fs_t* req;
-  if (lua_type(L, 3) == LUA_TTABLE) {
-    lua_getfield(L, 3, "dir");
-    if (lua_toboolean(L, -1)) flags |= UV_FS_SYMLINK_DIR;
-    lua_pop(L, 1);
-    lua_getfield(L, 3, "junction");
-    if (lua_toboolean(L, -1)) flags |= UV_FS_SYMLINK_JUNCTION;
-    lua_pop(L, 1);
+  // callback can be the 3rd parameter
+  if (luv_is_callable(L, 3) && lua_isnone(L, 4)) {
+    ref = luv_check_continuation(L, 3);
+  } else {
+    if (lua_type(L, 3) == LUA_TTABLE) {
+      lua_getfield(L, 3, "dir");
+      if (lua_toboolean(L, -1)) flags |= UV_FS_SYMLINK_DIR;
+      lua_pop(L, 1);
+      lua_getfield(L, 3, "junction");
+      if (lua_toboolean(L, -1)) flags |= UV_FS_SYMLINK_JUNCTION;
+      lua_pop(L, 1);
+    }
+    else if (lua_type(L, 3) == LUA_TNUMBER) {
+      flags = lua_tointeger(L, 3);
+    }
+    else if (!lua_isnoneornil(L, 3)) {
+      return luv_arg_type_error(L, 3, "table, integer, or nil expected, got %s");
+    }
+    ref = luv_check_continuation(L, 4);
   }
-  ref = luv_check_continuation(L, 4);
-  req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-
+  // ref the dest path so that we can print it in the error message
+  lua_pushvalue(L, 2);
+  ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
   FS_CALL(symlink, req, path, new_path, flags);
 }
 
@@ -729,7 +816,7 @@ static int luv_fs_readlink(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(readlink, req, path);
 }
@@ -739,7 +826,7 @@ static int luv_fs_realpath(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(realpath, req, path);
 }
@@ -751,7 +838,7 @@ static int luv_fs_chown(lua_State* L) {
   uv_uid_t uid = luaL_checkinteger(L, 2);
   uv_uid_t gid = luaL_checkinteger(L, 3);
   int ref = luv_check_continuation(L, 4);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(chown, req, path, uid, gid);
 }
@@ -762,7 +849,7 @@ static int luv_fs_fchown(lua_State* L) {
   uv_uid_t uid = luaL_checkinteger(L, 2);
   uv_uid_t gid = luaL_checkinteger(L, 3);
   int ref = luv_check_continuation(L, 4);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(fchown, req, file, uid, gid);
 }
@@ -774,7 +861,7 @@ static int luv_fs_lchown(lua_State* L) {
   uv_uid_t uid = luaL_checkinteger(L, 2);
   uv_uid_t gid = luaL_checkinteger(L, 3);
   int ref = luv_check_continuation(L, 4);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(lchown, req, path, uid, gid);
 }
@@ -787,32 +874,43 @@ static int luv_fs_copyfile(lua_State*L) {
   const char* new_path = luaL_checkstring(L, 2);
   int flags = 0, ref;
   uv_fs_t* req;
-  if (lua_type(L, 3) == LUA_TTABLE) {
-    lua_getfield(L, 3, "excl");
-    if (lua_toboolean(L, -1)) flags |= UV_FS_COPYFILE_EXCL;
-    lua_pop(L, 1);
+  // callback can be the 3rd parameter
+  if (luv_is_callable(L, 3) && lua_isnone(L, 4)) {
+    ref = luv_check_continuation(L, 3);
+  } else {
+    if (lua_type(L, 3) == LUA_TTABLE) {
+      lua_getfield(L, 3, "excl");
+      if (lua_toboolean(L, -1)) flags |= UV_FS_COPYFILE_EXCL;
+      lua_pop(L, 1);
 #if LUV_UV_VERSION_GEQ(1, 20, 0)
-    lua_getfield(L, 3, "ficlone");
-    if (lua_toboolean(L, -1)) flags |= UV_FS_COPYFILE_FICLONE;
-    lua_pop(L, 1);
-    lua_getfield(L, 3, "ficlone_force");
-    if (lua_toboolean(L, -1)) flags |= UV_FS_COPYFILE_FICLONE_FORCE;
-    lua_pop(L, 1);
+      lua_getfield(L, 3, "ficlone");
+      if (lua_toboolean(L, -1)) flags |= UV_FS_COPYFILE_FICLONE;
+      lua_pop(L, 1);
+      lua_getfield(L, 3, "ficlone_force");
+      if (lua_toboolean(L, -1)) flags |= UV_FS_COPYFILE_FICLONE_FORCE;
+      lua_pop(L, 1);
 #endif
+    }
+    else if (lua_type(L, 3) == LUA_TNUMBER) {
+      flags = lua_tointeger(L, 3);
+    }
+    else if (!lua_isnoneornil(L, 3)) {
+      return luv_arg_type_error(L, 3, "table, integer, or nil expected, got %s");
+    }
+    ref = luv_check_continuation(L, 4);
   }
-  else if (lua_type(L, 3) == LUA_TNUMBER) {
-    flags = lua_tointeger(L, 3);
-  }
-  ref = luv_check_continuation(L, 4);
-  req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
+  // ref the dest path so that we can print it in the error message
+  lua_pushvalue(L, 2);
+  ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
   FS_CALL(copyfile, req, path, new_path, flags);
 }
 #endif
 
 #if LUV_UV_VERSION_GEQ(1, 28, 0)
-static uv_dir_t* luv_check_dir(lua_State* L, int idx) {
-  uv_dir_t* dir = *(uv_dir_t**)luaL_checkudata(L, idx, "uv_dir");
+static luv_dir_t* luv_check_dir(lua_State* L, int idx) {
+  luv_dir_t* dir = (luv_dir_t*)luaL_checkudata(L, idx, "uv_dir");
   return dir;
 }
 
@@ -821,7 +919,7 @@ static int luv_fs_opendir(lua_State* L) {
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
   size_t nentries = luaL_optinteger(L, 3, 1);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
 
   //make data_ref to nentries
@@ -834,45 +932,44 @@ static int luv_fs_opendir(lua_State* L) {
 static int luv_fs_readdir(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   uv_fs_t *req;
-  uv_dir_t* dir = luv_check_dir(L, 1);
+  luv_dir_t* dir = luv_check_dir(L, 1);
   int ref = luv_check_continuation(L, 2);
 
-  req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(readdir, req, dir);
+  FS_CALL(readdir, req, dir->handle);
 }
 
 static int luv_fs_closedir(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
-  uv_dir_t* dir = luv_check_dir(L, 1);
+  luv_dir_t* dir = luv_check_dir(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t *req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+
+  luaL_unref(L, LUA_REGISTRYINDEX, dir->dirents_ref);
+  dir->dirents_ref = LUA_NOREF;
+
+  uv_fs_t *req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  lua_pushfstring(L, "uv_dir:%p", dir);
-  lua_pushnil(L);
-  lua_rawset(L, LUA_REGISTRYINDEX);
-  FS_CALL(closedir, req, dir);
+  FS_CALL(closedir, req, dir->handle);
 }
 
 static int luv_fs_dir_tostring(lua_State* L) {
-  uv_dir_t* dir = luv_check_dir(L, 1);
+  luv_dir_t* dir = luv_check_dir(L, 1);
   lua_pushfstring(L, "uv_dir_t: %p", dir);
   return 1;
 }
 
 static int luv_fs_dir_gc(lua_State* L) {
-  uv_dir_t* dir = luv_check_dir(L, 1);
-  lua_pushfstring(L, "uv_dir:%p", dir);
-  lua_rawget(L, LUA_REGISTRYINDEX);
-  if (!lua_isnil(L, -1)) {
+  luv_dir_t* dir = luv_check_dir(L, 1);
+  if (dir->dirents_ref != LUA_NOREF) {
     uv_fs_t req;
     luv_ctx_t* ctx = luv_context(L);
 
-    uv_fs_closedir(ctx->loop, &req, dir, NULL);
+    luaL_unref(L, LUA_REGISTRYINDEX, dir->dirents_ref);
+    dir->dirents_ref = LUA_NOREF;
+
+    uv_fs_closedir(ctx->loop, &req, dir->handle, NULL);
     uv_fs_req_cleanup(&req);
-    lua_pushfstring(L, "uv_dir:%p", dir);
-    lua_pushnil(L);
-    lua_rawset(L, LUA_REGISTRYINDEX);
   }
   lua_pop(L, 1);
 
@@ -885,7 +982,7 @@ static int luv_fs_statfs(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
-  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(statfs, req, path);
 }
